/**************************************************************************
* iptrk - ip tracker
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia.
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* iptrk query
*
**************************************************************************/
#include "defs.h"
#include "lib.p"

typedef struct
{
  U_32    addr;
  voldr_t voldr;
} addr_vol_t;

typedef struct
{
  U_32    addr;  
  treap_t *a_ll; 
} addr_vol_ass_t;

PRIVATE int     range_type;
PRIVATE U_32    start_addr, end_addr = 0xffffffff;
PRIVATE treap_t *query_bt,
                *addr_vol_bt,
                *our_net_ip_bt;
PRIVATE FILE    *file_fp;
PRIVATE time_t  start_time,
                end_time;

/**************************************************************************
* load routines
**************************************************************************/
PRIVATE int load_file_port(treap_t **t)
{
  char   str[STR_LEN + 1];
  U_16   port_s, port_d;
  U_64   vol;
  port_t entry, *p;

  int find_cmp_func(void *n)
  {
    return cmp_32(entry.port.key, ((port_t *)n)->port.key);
  }

  for(;;)
  {
    if(feof(file_fp))
      return 0;

    if(fgets(str, STR_LEN, file_fp) == NULL)
      return 0;

    if(str[0] == '\n')
      return 1;

    if(sscanf(str, "%04hx %04hx %llx\n", &port_s, &port_d, &vol) != 3)
      return 0;

    entry.port.port.s = port_s;
    entry.port.port.d = port_d;
    p = treap_find(t, find_cmp_func);
    if(p == NULL)
    {
      entry.vol = 0;
      treap_append(t, &entry, sizeof(port_t), port_append_cmp_func);
      p = treap_find(t, find_cmp_func);
    }

    p->vol += vol;
  }

  return 1;
}

PRIVATE int load_file_iptrk(treap_t **bt)
{
  char    str[STR_LEN + 1];
  U_32    src = 0, dst = 0;
  U_64    vol = 0;
  U_32    c_ts, u_ts;
  iptrk_t entry, *p;

  int find_cmp_func(void *n)
  {
    return cmp_64(entry.addr.key, ((iptrk_t *)n)->addr.key);
  }

  for(;;)
  {
    if(feof(file_fp))
      return 1;

    if(fgets(str, STR_LEN, file_fp) == NULL)
      return 1;

    if(sscanf(str, "%lx %lx %llx %lx %lx",
      &src, &dst, &vol, &c_ts, &u_ts) != 5)
    {
      return 0;
    }

    entry.addr.addr.s = src;
    entry.addr.addr.d = dst;
    p = treap_find(bt, find_cmp_func);
    if(p == NULL)
    {
      entry.oth_port_ll = NULL;
      entry.tcp_port_ll = NULL;
      entry.udp_port_ll = NULL;
      entry.vol = 0;
      entry.created = timer;
      entry.last_updated = 0;
      treap_append(bt, &entry, sizeof(iptrk_t), iptrk_append_cmp_func);
      p = treap_find(bt, find_cmp_func);
    }

    p->vol += vol;
    if(c_ts < p->created)
      p->created = c_ts;
    if(u_ts > p->last_updated)
      p->last_updated = u_ts;

    if(fgets(str, STR_LEN, file_fp) == NULL)
      return 0;
    if(strncmp(str, "O\n", 2))
      return 0;
    if(load_file_port(&(p->oth_port_ll)) == 0)
      return 0;

    if(fgets(str, STR_LEN, file_fp) == NULL)
      return 0;
    if(strncmp(str, "T\n", 2))
      return 0;
    if(load_file_port(&(p->tcp_port_ll)) == 0)
      return 0;

    if(fgets(str, STR_LEN, file_fp) == NULL)
      return 0;
    if(strncmp(str, "U\n", 2))
      return 0;
    if(load_file_port(&(p->udp_port_ll)) == 0)
      return 0;
  }

  return 1;
}

PUBLIC int load_file(treap_t **bt, char *filename)
{
  file_fp = fopen(filename, "r");
  if(file_fp == NULL)
  {
    log_e(__func__, "::fopen()");
    return 0;
  }

  if(!load_file_iptrk(bt))
    log_m('W', __func__, "- corrupted file");

  fclose(file_fp);

  return 1;
}

/**************************************************************************
*
**************************************************************************/
PUBLIC void load_directory(treap_t **bt, char *path)
{
  DIR           *dp;
  struct dirent *dir_p;
  struct stat   stat_buf;
  char          str[STR_LEN + 1], *s;
  time_t        tm;

  dp = opendir(path);
  if(dp == NULL)
  {
    log_e(__func__, "::opendir()");
    exit(-1);
  }

  strcpy(str, path);
  strcat(str, "/");
  s = str + strlen(str);

  for(;;)
  {
     dir_p = readdir(dp);
     if(dir_p == NULL)
       break;

     strcpy(s, dir_p->d_name);
     if(stat(str, &stat_buf) == -1)
     {
       log_e(__func__, "::stat()");
       exit(-1);
     }

     if(S_ISDIR(stat_buf.st_mode))
       continue;

     tm = datestr2time(dir_p->d_name + 3);
     if(tm < start_time || tm > end_time)
       continue;

     load_file(bt, str);
  }

  closedir(dp);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE inline int test_range(U_32 a)
{
  if(a >= start_addr && a <= end_addr)
    return 1;

  return 0;
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void print_report_header(void)
{
  printf(
    "%-15s %20s %20s %20s %20s %20s\n"
    "--\n",
    "ADDRESS", "I", "O", "u", "i", "o");
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void dump_db(void)
{
  struct in_addr in;
  char           str[STR_LEN + 1],
                 str2[STR_LEN + 1],
                 str3[STR_LEN + 1],
                 str4[STR_LEN + 1];
  U_32           a;

  void w_port(treap_t *t)
  {
    if(t != NULL)
    {
      port_t *e;

      w_port(t->l);

      e = TREAP_ENTRY(t);
      if(!e->port.port.s)
        strcpy(str2, "*");
      else
        sprintf(str2, "%hu", e->port.port.s);

      if(!e->port.port.d)
        strcpy(str3, "*");
      else
        sprintf(str3, "%hu", e->port.port.d);

      print_vol_u64(str4, e->vol);
      printf("\t\t%-5s %-5s %s\n", str2, str3, str4);

      w_port(t->r);
    }
  }

  void w_tree(treap_t *t)
  {
    if(t != NULL)
    {
      iptrk_t *n;

      w_tree(t->l);

      n = TREAP_ENTRY(t);

      if(range_type == 0)
        a = n->addr.addr.s;
      else
        a = n->addr.addr.d;

      if(test_range(a))
      {
        in.s_addr = htonl(n->addr.addr.s);
        strcpy(str, inet_ntoa(in));
        in.s_addr = htonl(n->addr.addr.d);

        format_date_entry(n->created, str2);
        format_date_entry(n->last_updated, str3);

        print_vol_u64(str4, n->vol);
        printf("o %-15s -> %-15s\t%15s (%s) (%s) %c\n",
          str, inet_ntoa(in), str4, str2, str3,
          n->created == n->last_updated ? '*' : ' ');

        printf("\n\tTCP:\n");
        w_port(n->tcp_port_ll);

        printf("\n\tUDP:\n");
        w_port(n->udp_port_ll);

        printf("\n\tOTH:\n");
        w_port(n->oth_port_ll);
        puts("\n\n");
      }

      w_tree(t->r);
    }
  }

  w_tree(query_bt);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void update_addr_vol(treap_t **bt, U_32 a_addr, voldr_t *a_v)
{
  addr_vol_t entry, *p;

  int find_cf(void *n)
  {
    return cmp_32(a_addr, ((addr_vol_t *)n)->addr);
  }

  int append_cf(void *n1, void *n2)
  {
    return cmp_32(((addr_vol_t *)n1)->addr, ((addr_vol_t *)n2)->addr);
  }

  p = treap_find(bt, find_cf);
  if(p == NULL)
  {
    entry.addr = a_addr;
    memcpy(&(entry.voldr), a_v, sizeof(voldr_t));
    treap_append(bt, &entry, sizeof(addr_vol_t), append_cf);

    p = treap_find(bt, find_cf);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }
  else
    voldr_acc(&(p->voldr), a_v);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void update_addr_vol_ass(U_32 addr, U_32 a_addr, voldr_t *a_v)
{
  addr_vol_ass_t *p, entry;

  int find_cmp_func(void *n)
  {
    return cmp_32(entry.addr, ((addr_vol_ass_t *)n)->addr);
  }

  int append_cmp_func(void *n1, void *n2)
  {
    return cmp_32(((addr_vol_ass_t *)n1)->addr, ((addr_vol_ass_t *)n2)->addr);
  }

  entry.addr = addr;
  p = treap_find(&addr_vol_bt, find_cmp_func);
  if(p == NULL)
  {
    entry.a_ll = NULL;
    treap_append(&addr_vol_bt, &entry, sizeof(addr_vol_ass_t), append_cmp_func);

    p = treap_find(&addr_vol_bt, find_cmp_func);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }

  update_addr_vol(&(p->a_ll), a_addr, a_v);
}

/**************************************************************************
* address -> address associative report
**************************************************************************/
PRIVATE void compile_addr_vol_ass(void)
{
  addr_t  key_addr;
  voldr_t v;

  void search_tree(treap_t *t)
  {
    iptrk_t *n;

    if(t == NULL)
      return;

    search_tree(t->l);

    n = TREAP_ENTRY(t);
    if(key_addr.addr == n->addr.addr.s)
    {
      cast_volume_log(n->addr.addr.s, n->addr.addr.d, &key_addr, n->vol, &v);
      update_addr_vol_ass(n->addr.addr.s, n->addr.addr.d, &v);
    }
    else
      if(key_addr.addr == n->addr.addr.d)
      {
        cast_volume_log(n->addr.addr.s, n->addr.addr.d, &key_addr, n->vol, &v);
        update_addr_vol_ass(n->addr.addr.d, n->addr.addr.s, &v);
      }

    search_tree(t->r);
  }

  void w_local_ips(treap_t *t)
  {
    if(t == NULL)
      return;

    w_local_ips(t->l);

    key_addr.addr = *((U_32 *)TREAP_ENTRY(t));
    if(test_range(key_addr.addr))
      search_tree(query_bt);

    w_local_ips(t->r);
  }

  key_addr.mask = 0xffffffff;
  gather_local_ips(query_bt, &our_net_ip_bt);
  w_local_ips(our_net_ip_bt);
}

PRIVATE void print_addr_vol_ass(void)
{
  struct in_addr in;
  char           str1[STR_LEN + 1],
                 str2[STR_LEN + 1],
                 str3[STR_LEN + 1],
                 str4[STR_LEN + 1],
                 str5[STR_LEN + 1];

  void w_tree_0(treap_t *t)
  {
    addr_vol_t *n;

    if(t == NULL)
      return;

    w_tree_0(t->l);

    n = TREAP_ENTRY(t);
    in.s_addr = htonl(n->addr);

    print_vol_u64(str1, n->voldr.I);
    print_vol_u64(str2, n->voldr.O);
    print_vol_u64(str3, n->voldr.u);
    print_vol_u64(str4, n->voldr.i);
    print_vol_u64(str5, n->voldr.o);
    printf("%-15s %20s %20s %20s %20s %20s\n",
      inet_ntoa(in), str1, str2, str3, str4, str5);

    w_tree_0(t->r);
  }

  void w_tree(treap_t *t)
  {
    addr_vol_ass_t *n;

    if(t == NULL)
      return;

    w_tree(t->l);

    n = TREAP_ENTRY(t);
    in.s_addr = htonl(n->addr);
    printf("o %-15s\n", inet_ntoa(in));
    w_tree_0(n->a_ll);
    putchar('\n');

    w_tree(t->r);
  }

  w_tree(addr_vol_bt);
}

/**************************************************************************
* address volume report (used for billing etc)
**************************************************************************/
PRIVATE void compile_addr_vol(void)
{
  addr_t  kaddr, *kaddr_p = &kaddr;
  voldr_t v;
  ll_t    *ll;

  void s_tree(treap_t *t)
  {
    iptrk_t *n;

    if(t == NULL)
      return;

    s_tree(t->l);

    n = TREAP_ENTRY(t);
    if(kaddr_p->addr == (n->addr.addr.s & kaddr_p->mask))
    {
      cast_volume_log(n->addr.addr.s, n->addr.addr.d, kaddr_p, n->vol, &v);
      update_addr_vol(&addr_vol_bt, kaddr_p->addr, &v);
    }
    else
      if(kaddr_p->addr == (n->addr.addr.d & kaddr_p->mask))
      {
        cast_volume_log(n->addr.addr.s, n->addr.addr.d, kaddr_p, n->vol, &v);
        update_addr_vol(&addr_vol_bt, kaddr_p->addr, &v);
      }

    s_tree(t->r);
  }

  void w_local_ips(treap_t *t)
  {
    if(t == NULL)
      return;

    w_local_ips(t->l);

    kaddr_p->addr = *((U_32 *)TREAP_ENTRY(t));
    if(test_range(kaddr_p->addr))
      s_tree(query_bt);

    w_local_ips(t->r);
  }

  gather_local_ips(query_bt, &our_net_ip_bt);
  kaddr_p->mask = 0xffffffff;
  w_local_ips(our_net_ip_bt);

  ll = boot.extra_report_ll;
  for(;;)
  {
    kaddr_p = ll_read_entry(&ll);
    if(kaddr_p == NULL)
      break;

    if(test_range(kaddr_p->addr))
      s_tree(query_bt);
  }
}

PRIVATE void print_addr_vol(void)
{
  struct in_addr in;
  char           str1[STR_LEN + 1],
                 str2[STR_LEN + 1],
                 str3[STR_LEN + 1],
                 str4[STR_LEN + 1],
                 str5[STR_LEN + 1];

  void w_tree(treap_t *t)
  {
    addr_vol_t *n;

    if(t == NULL)
      return;

    w_tree(t->l);

    n = TREAP_ENTRY(t);
    in.s_addr = htonl(n->addr);

    print_vol_u64(str1, n->voldr.I);
    print_vol_u64(str2, n->voldr.O);
    print_vol_u64(str3, n->voldr.u);
    print_vol_u64(str4, n->voldr.i);
    print_vol_u64(str5, n->voldr.o);
    printf("%-15s %20s %20s %20s %20s %20s\n",
      inet_ntoa(in), str1, str2, str3, str4, str5);

    w_tree(t->r);
  }

  w_tree(addr_vol_bt);
}

/**************************************************************************
*
**************************************************************************/
int main(int argc, char *argv[])
{
  int            x, mode = 0;
  struct stat    stat_buf;
  char           str[STR_LEN + 1],
                 db_path[STR_LEN + 1] = IPTRK_DB_WATCH_PATHC;
  struct in_addr in;

  welcome();
  timer = end_time = time(NULL);
  for(x = 0; x < argc; x++)
  {
    if(argv[x][0] == '-')
    {
      switch(argv[x][1])
      {
        case 'h':
          printf(
            "Usage: iptrkq <options>\n"
            " -h           Display this help info\n"
            " -f <db path> Database (file/directory) to interrogate\n"
            " -m <mode>    Mode (0 = dump, 1 = report 1, 2 = report 2)\n"
            " -S           Test range against Src Address (only App. to dump)\n"
            " -D           Test range against Dst Address (ditto)\n"
            " -s <address> Start Address\n"
            " -e <address> End Address\n"
            " -t <time>    Start time (Default = Epoch)\n"
            " -n <time>    End time   (Default = Current time)\n"
            "    (Nb/ Time format can be either YYYY:MM:DD:HH:MM or HH:MM)\n"
            "\n"
          );
          return 0;

        case 'f':
          strncpy2(db_path, parse_cmdline(argc, argv, x), STR_LEN);
          break;

        case 'm':
          mode = atoi(parse_cmdline(argc, argv, x));
          break;

        case 'D':
          range_type = 1;
          break;

        case 's':
          inet_aton(parse_cmdline(argc, argv, x), &in);
          start_addr = ntohl(in.s_addr);
          break;

        case 'e':
          inet_aton(parse_cmdline(argc, argv, x), &in);
          end_addr = ntohl(in.s_addr);
          break;

        case 't':
          start_time = datestr2time(parse_cmdline(argc, argv, x));
          break;

        case 'n':
          end_time = datestr2time(parse_cmdline(argc, argv, x));
          break;
      }
    }
  }

  timer = time(NULL);
  if(stat(db_path, &stat_buf) == -1)
  {
    log_e(__func__, "::stat()");
    exit(-1);
  }

  read_boot_conf();
  sprintf(str, " loading db %s", db_path);
  log_m('I', NULL, str);
  if(S_ISDIR(stat_buf.st_mode))
    load_directory(&query_bt, db_path);
  else
    if(!load_file(&query_bt, db_path))
      exit(-1);

  timer = time(NULL);
  time_st = localtime(&timer);
  memcpy(&old_time_st, time_st, sizeof(struct tm));

  in.s_addr = htonl(start_addr);
  strcpy(str, inet_ntoa(in));
  in.s_addr = htonl(end_addr);
  printf("\n"
         " last updated: %lu min ago\n"
         "   test range: %s (Dump only)\n"
         "address range: %s -> %s\n",
    (timer - stat_buf.st_ctime) / 60,
    range_type ? "Destination" : "Source",
    str, inet_ntoa(in));
  format_date_entry(start_time, str);
  printf(
         "   start time: %s\n", str);
  format_date_entry(end_time, str);
  printf(
         "     end time: %s\n\n", str);

  printf("--\n");
  switch(mode)
  {
    case 0:
      dump_db();
      break;

    case 1:
      print_report_header();
      compile_addr_vol_ass();
      print_addr_vol_ass();
      break;

    case 2:
      print_report_header();
      compile_addr_vol();
      print_addr_vol();
      break;

    default:
      printf("E unknown mode.\n");
      break;
  }

  printf("--\nEND\n");

  return 0;
}
